/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File ViewArea.h */
/* This is the implementation file of the ViewArea widget. */

#include <stdio.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "ViewAreaP.h"

#define DEFAULT_VIEW_WIDTH 512
#define DEFAULT_VIEW_HEIGHT 512

#define offset(field) XtOffsetOf(ViewAreaRec,field)
static XtResource resources[] = {
  {XtNcallback,XtCCallback,XtRCallback,sizeof(XtPointer),
   offset(viewArea.callback),XtRCallback,NULL},
  {XtNviewWidth,XtCViewWidth,XtRDimension,sizeof(Dimension),
   offset(viewArea.view_width),
   XtRImmediate,(XtPointer)DEFAULT_VIEW_WIDTH},
  {XtNviewHeight,XtCViewHeight,XtRDimension,sizeof(Dimension),
   offset(viewArea.view_height),
   XtRImmediate,(XtPointer)DEFAULT_VIEW_HEIGHT}
};
#undef offset

static ViewAreaNotifyInfo info;

static void Notify(Widget,XEvent *,String *,Cardinal *);

static XtActionsRec actions[] = {
  {"notify",Notify}
};

static char defaultTranslations[] =
"<Btn1Down>: notify()\n\
<Btn3Down>: notify()";

/* The method prototypes */
static void Initialize(Widget,Widget,ArgList,Cardinal *);
static void Redisplay(Widget,XEvent *,Region);
static void Destroy(Widget);
static void Resize(Widget);
static Boolean SetValues(Widget,Widget,Widget,ArgList,Cardinal *);
static XtGeometryResult QueryGeometry(Widget,XtWidgetGeometry *,
				      XtWidgetGeometry *);

#define SuperClass ((CoreWidgetClass)&coreClassRec)

ViewAreaClassRec viewAreaClassRec = {
  { /* core */
    (WidgetClass)SuperClass,            /* superclass */
    "ViewArea",                         /* class_name */
    sizeof(ViewAreaRec),                /* widget_size */
    NULL,                               /* class_initialize */
    NULL,                               /* class_part_initialize */
    False,                              /* class_inited */
    Initialize,                         /* initialize */
    NULL,                               /* initialize_hook */
    XtInheritRealize,                   /* realize */
    actions,                            /* actions */
    XtNumber(actions),                  /* num_actions */
    resources,                          /* resources */
    XtNumber(resources),                /* num_resources */
    NULLQUARK,                          /* xrm_class */
    True,                               /* compress_motion */
    True,                               /* compress_exposure */
    True,                               /* compress_interleave */
    False,                              /* visible_interest */
    Destroy,                            /* destroy */
    Resize,                             /* resize */
    Redisplay,                          /* expose */
    SetValues,                          /* set_values */
    NULL,                               /* set_values_hook */
    XtInheritSetValuesAlmost,           /* set_values_almost */
    NULL,                               /* get_values_hook */
    NULL,                               /* accept_focus */
    XtVersion,                          /* version */
    NULL,                               /* callback_private */
    defaultTranslations,                /* tm_table */
    QueryGeometry,                      /* query_geometry */
    XtInheritDisplayAccelerator,        /* display_accelerator */
    NULL                                /* extension */
  },
  { /* viewArea */
    0                                   /* dummy */
  }
};

WidgetClass viewAreaWidgetClass = (WidgetClass)&viewAreaClassRec;

static void
Initialize(Widget request,Widget init,ArgList args,Cardinal *num_args)
{
  ViewAreaWidget w = (ViewAreaWidget)init; /* I love C! */
  if (w->viewArea.view_width<1) {
    XtWarning("ViewArea: viewWidth is too small (using 1).");
    w->viewArea.view_width=1;
  }
  if (w->viewArea.view_height<1) {
    XtWarning("ViewArea: viewHeight is too small (using 1).");
    w->viewArea.view_height=1;
  }
  w->core.width=w->viewArea.view_width;
  w->core.height=w->viewArea.view_height;
#if 0
  (*XtClass(w)->core_class.resize)(Widget(w)); /* ??? */
#endif
  w->viewArea.view_pixmap=XCreatePixmap(XtDisplay(init), /* XXX */
					DefaultRootWindow(XtDisplay(init)),
					w->viewArea.view_width,
					w->viewArea.view_height,
					DefaultDepthOfScreen(XtScreen(init)));
}

Pixmap
ViewAreaGetViewPixmap(Widget w0)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  return w->viewArea.view_pixmap;
}

void
ViewAreaRedrawRegion(Widget w0,int x1,int y1,int x2,int y2)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  int i,j;
  if (!XtIsRealized(w0)) return;
  if ((x2<=x1)||(y2<=y1)) return;
  if (x1<0) x1=0;
  if (y1<0) y1=0;
  if (x2>=w->viewArea.view_width)
    x2=w->viewArea.view_width-1;
  if (y2>=w->viewArea.view_height)
    y2=w->viewArea.view_height-1;
  XCopyArea(XtDisplay(w0),w->viewArea.view_pixmap,XtWindow(w0),
	    DefaultGCOfScreen(XtScreen(w0)),x1,y1,x2-x1,y2-y1,x1,y1);
}

static void
Redisplay(Widget w0,XEvent *event,Region region)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  ViewAreaRedrawRegion(w0,0,0,w->viewArea.view_width,w->viewArea.view_height);
}

static void
Destroy(Widget w0)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  XFreePixmap(XtDisplay(w0),w->viewArea.view_pixmap);
}

static void
Resize(Widget w0)
{
  /* HOW DARE YOU RESIZE MY WIDGET YOU STUPID X TOOLKIT */
}

static Boolean
SetValues(Widget current,Widget request,Widget new,ArgList args,
	  Cardinal *num_args)
{
  ViewAreaWidget curw = (ViewAreaWidget)current;
  ViewAreaWidget neww = (ViewAreaWidget)new;
  if (neww->viewArea.view_width!=
      curw->viewArea.view_width) {
    XtWarning("ViewArea: viewWidth is read-only.");
    neww->viewArea.view_width
      =curw->viewArea.view_width;
  }
  if (neww->viewArea.view_height!=
      curw->viewArea.view_height) {
    XtWarning("ViewArea: viewHeight is read-only.");
    neww->viewArea.view_height
      =curw->viewArea.view_height;
  }
  return True;
}

static XtGeometryResult
QueryGeometry(Widget w0,XtWidgetGeometry *proposed,XtWidgetGeometry *answer)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  answer->request_mode=CWWidth|CWHeight;
  answer->width=w->viewArea.view_width;
  answer->height=w->viewArea.view_height;
  if (((proposed->request_mode&(CWWidth|CWHeight))==(CWWidth|CWHeight))&&
      (proposed->width==answer->width)&&(proposed->height==answer->height))
    return XtGeometryYes;
  else if ((answer->width==w->core.width)&&(answer->height==w->core.height))
    return XtGeometryNo;
  else return XtGeometryAlmost;
}

static void
Notify(Widget w0,XEvent *event,String *params,Cardinal *num_params)
{
  ViewAreaWidget w = (ViewAreaWidget)w0;
  if ((event->type==ButtonPress)||(event->type==ButtonRelease)) {
    info.x=event->xbutton.x;
    info.y=event->xbutton.y;
  } else if ((event->type==KeyPress)||(event->type==KeyRelease)) {
    info.x=event->xkey.x;
    info.y=event->xkey.y;
  } else {
    info.x=0; /* Actually, they are undefined */
    info.y=0;
  }
  info.event=event;
  info.params=params;
  info.num_params=num_params;
  XtCallCallbacks(w0,XtNcallback,&info);
}

/* EOF */
