/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File fifo.c */
/* This file contains the functions that handle the control fifo,
 * through which the client threads pass messages to the server
 * thread. */


#include <stdio.h>
#include <pthread.h>

#include "lgdlib.h"
#include "fifo.h"

msg_fifo_t *control_queue_head = NULL;
msg_fifo_t *control_queue_tail = NULL;
pthread_mutex_t msg_fifo_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t msg_fifo_cond = PTHREAD_COND_INITIALIZER;

long int nextserial;

void
append_control_msg(msg_fifo_t *msg)
     /* Add a message to the control fifo.  This is used by the client
      * threads to place a request.  This function fills in the
      * treated, next, serial and timestamp entries of the msg
      * structure, and also the ready entry of the resp structure if
      * it exists. */
{
  /* Fill in missing entries */
  msg->treated=0;
  msg->next=NULL;
  if (msg->response) {
    msg->response->ready_stat=0;
    pthread_cond_init(&msg->response->ready_cond,NULL);
    pthread_mutex_init(&msg->response->ready_mutex,NULL);
  }
  /* Now we lock the control fifo. */
  pthread_mutex_lock(&msg_fifo_mutex);
  msg->serial=nextserial++;
  gettimeofday(&msg->timestamp,NULL);
  if (control_queue_tail==NULL) {
    control_queue_tail=msg;
    control_queue_head=msg;
  } else {
    control_queue_tail->next=msg;
    control_queue_tail=msg;
  }
  /* Wake up the server thread */
  /* DO NOT USE pthread_cond_signal() HERE.  It will work in case
   * there is just one server thread, but if there is more than one
   * waiting, you have a pretty deadlock alert... */
  pthread_cond_broadcast(&msg_fifo_cond);
  /* Unlock the control fifo */
  pthread_mutex_unlock(&msg_fifo_mutex);
}

msg_fifo_t *
wait_control_msg(void)
     /* Return first message of the control fifo and remove it
      * therefrom.  This is used by the server threads to know what to
      * do.  If the control fifo is empty, wait for a message to
      * arrive. */
{
  msg_fifo_t *ret;
  /* Acquire the control fifo */
  pthread_mutex_lock(&msg_fifo_mutex);
  /* Wait for the condition to be signalled */
  while (control_queue_head==NULL)
    pthread_cond_wait(&msg_fifo_cond,&msg_fifo_mutex);
  /* Note that this atomically unlocks the mutex and waits on the
   * condition.  We don't have a deadlock here! */
  /* Remove the message from the queue */
  ret=control_queue_head;
  control_queue_head=ret->next;
  if (control_queue_head==NULL) /* We removed the only message */
    control_queue_tail=NULL;
  /* Now unlock the control fifo */
  pthread_mutex_unlock(&msg_fifo_mutex);
  return ret;
}
