/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File fifo.h */
/* This file contains the prototypes and declarations for the control
 * fifo handling routines, contained in fifo.c */

/* Avoid multiple inclusions. */
#ifndef _FIFO_H
#define _FIFO_H

#include <sys/time.h>
#include <pthread.h>

/* The possible message types */
typedef enum {
  msg_noop, /* Do nothing */
  msg_die,  /* Die */
  msg_readmap,
  msg_write_chunk,
  msg_write_square,
  msg_sync_map,
  msg_move_hero,
  msg_whereis_hero
} msg_type_t;

struct msg_response_s;

/* The control fifo is a chained list of messages to be treated by the
 * server threads.  The client threads apprend messages at the end of
 * the list, and the server thread(s) gobble them at the beginning. */
typedef struct msg_fifo_s {
  /* next points to the next message in the chain */
  struct msg_fifo_s *next;
  /* treated indicates whether the request has been treated by the
   * server.  The client should set it to zero before submitting the
   * request and the server should put it to one when the request has
   * been treated.  This is important in the case where the structure
   * is being stored on the client's stack: it should not modify it
   * until the server has set this bit. */
  char treated;
  /* in_heap is set to one by the client when the structure has been
   * malloc()ated on the heap and the server should free() it when it
   * has finished playing with it. */
  char in_heap;
  /* resp points to the place where the server should store the
   * response.  It will set resp->ready when the response is actually
   * ready.  If the client does not want to hear a response from the
   * server, it should initialize resp to NULL. */
  struct msg_response_s *response;
  /* Serial number of the message */
  long int serial;
  /* This is the timestamp of message submission.  As of yet, this is
   * unused and may disappear. */
  struct timeval timestamp;
  msg_type_t type; /* Type of request */
  union { /* Message parameters - according to type */
    int foobar; /* Compiler may not like an empty union... */
    struct { long int x,y; } readmap;
    struct { long int x,y; u_int32_t v; } write_chunk;
    struct { long int x,y; u_int32_t v; } write_square;
    struct { long int dx,dy; } move_hero;
  } param;
} msg_fifo_t;

/* First message in control fifo */
extern msg_fifo_t *control_queue_head; /* NULL if no messages */
/* Last message in control fifo */
extern msg_fifo_t *control_queue_tail; /* NULL if no messages */

/* The following mutex protects the control fifo.  While it is locked,
 * the control fifo is being used by some thread and should not be
 * touched. */

extern pthread_mutex_t msg_fifo_mutex;

/* The following condition is used to wake up the server thread(s)
 * when a request is available. */

extern pthread_cond_t msg_fifo_cond;

/* The possible error codes */
typedef enum {
  msg_err_ok, /* Request was successfully carried out */
  msg_err_unknown, /* Unknown message type received */
  msg_err_generic, /* Generic (unspecified) error */
  msg_err_blocked /* ``Blocked'' */
} msg_errcode_t;

/* This is the structure in which the server thread stores the reponse
 * to the client requests. */
typedef struct msg_response_s {
  char ready_stat; /* Set to 1 when response is available */
  pthread_cond_t ready_cond; /* Activated when response is available */
  pthread_mutex_t ready_mutex; /* Protects the above variables */
  msg_errcode_t errcode; /* 0 if request has been _successfully_ completed */
  long int ques_serial; /* Serial number of the question being answered */
  union {
    int foobar;
    struct { u_int32_t val; char hero; } readmap;
    struct { long int x,y; } whereis_hero;
  } ret;
} msg_response_t;

/* Prototypes */

void append_control_msg(msg_fifo_t *msg);

msg_fifo_t *wait_control_msg(void);

#endif /* not _FIFO_H */
