/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File input.c */
/* This file contains the input handling thread. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>

#include <guile/gh.h>

#include "lgd0.h"
#include "lgdlib.h"
#include "fifo.h"

#include "ViewArea.h"

#include "input.h"
#include "output.h"
#include "legendes.h"

pthread_t input_thread;

static char *fallback_resources[] = {
  "*input: True",
  "*background: Grey51",
  "*foreground: White",
  "*groundBox.background: DarkSlateGrey",
  "*groundBox.cuteLabel.foreground: Blue",
  "*groundBox.quitButton.label: Quit",
  "*groundBox.submitButton.label: (enter)",
  "*groundBox.viewArea.translations: #override \\\n\
     <Key>Left: notify(move-west) \\n\\\n\
     <Key>Right: notify(move-east) \\n\\\n\
     <Key>Up: notify(move-north) \\n\\\n\
     <Key>Down: notify(move-south)",
  NULL
};

Widget toplevel,ground_box,cute_label,quit_button;
Widget command_line,submit_button;
Widget view_area;

static XtAppContext app_context;

Display *dpy;
Screen *scn;
Window win;

Pixmap disp_pm;

static SCM
cmd_move_west(void)
{
  req_move_hero(NULL,-1,0);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_move_east(void)
{
  req_move_hero(NULL,1,0);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_move_north(void)
{
  req_move_hero(NULL,0,-1);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_move_south(void)
{
  req_move_hero(NULL,0,1);
  return SCM_UNSPECIFIED;
}

static void
command(char *cmd)
{
  gh_eval_str(cmd);
}

static void
quit_callback(Widget what,XtPointer client_data,XtPointer call_data)
{
  exit(0);
}

static void
submit_callback(Widget what,XtPointer client_data,XtPointer call_data)
{
  char *s,*t;
  XtVaGetValues(command_line,XtNstring,&s,NULL);
  t=lgd_malloc(1,strlen(s)+1);
  strcpy(t,s);
  XtVaSetValues(command_line,XtNstring,"",NULL);
  command(t);
  lgd_free(t);
}

static void
view_callback(Widget what,XtPointer client_data,XtPointer call_data)
{
  ViewAreaNotifyInfo *info = (ViewAreaNotifyInfo *)call_data;
  if (*info->num_params>=1) {
    char *t;
    t=lgd_malloc(1,strlen(info->params[0])+3);
    t[0]='(';
    strcpy(t+1,info->params[0]);
    t[strlen(info->params[0])+1]=')';
    t[strlen(info->params[0])+2]=0;
    command(t);
    lgd_free(t);
  } else command("()");
}

void
init_20_input(void)
{
  /* Initialize thread support (Xlib and X toolkit) */
  if (XInitThreads()==0)
    lgd_fatal("This Xlib does not support threads!");
  if (XtToolkitThreadInitialize()!=True)
    lgd_fatal("This Xt does not support threads!");
  /* Use the default language procedure (?) */
  XtSetLanguageProc(NULL,(XtLanguageProc)NULL,NULL);
  /* Create the widgets */
  toplevel = XtVaAppInitialize(&app_context,APP_CLASS,
			       NULL,0,&argc,argv,
			       fallback_resources,
			       NULL);
  ground_box = XtVaCreateManagedWidget("groundBox",boxWidgetClass,
				       toplevel,
				       NULL);
  cute_label = XtVaCreateManagedWidget("cuteLabel",labelWidgetClass,
				       ground_box,
				       XtNlabel,GAMENAME,
				       NULL);
  quit_button = XtVaCreateManagedWidget("quitButton",commandWidgetClass,
					ground_box,
					NULL);
  command_line = XtVaCreateManagedWidget("commandLine",asciiTextWidgetClass,
					 ground_box,
					 XtNeditType,XawtextEdit,
					 NULL);
  submit_button = XtVaCreateManagedWidget("submitButton",commandWidgetClass,
					  ground_box,
					  NULL);
  view_area = XtVaCreateManagedWidget("viewArea",viewAreaWidgetClass,
				      ground_box,
				      XtNviewWidth,WIDTH*ICONSIZE,
				      XtNviewHeight,HEIGHT*ICONSIZE,
				      NULL);
  /* Specify the callbacks */
  XtAddCallback(quit_button,XtNcallback,quit_callback,NULL);
  XtAddCallback(submit_button,XtNcallback,submit_callback,NULL);
  XtAddCallback(view_area,XtNcallback,view_callback,NULL);
  /* Get the ball rolling */
  XtRealizeWidget(toplevel);
  /* Set the important variables */
  dpy=XtDisplay(view_area);
  scn=XtScreen(view_area);
  win=XtWindow(view_area);
  disp_pm=ViewAreaGetViewPixmap(view_area);
  /* Install the various functions of the scheme interpreter */
  gh_new_procedure0_0("move-west",cmd_move_west);
  gh_new_procedure0_0("move-east",cmd_move_east);
  gh_new_procedure0_0("move-north",cmd_move_north);
  gh_new_procedure0_0("move-south",cmd_move_south);
}

void
demand_redraw(void)
     /* Send an X Expose event so that the ViewArea widget will redraw
      * itself.  This is used by some other thread (the output thread,
      * presumably).  _Not_ by the input thread (of course). */
{
  XEvent evt;
  evt.type=Expose;
  evt.xexpose.x=0;
  evt.xexpose.y=0;
  evt.xexpose.width=512;
  evt.xexpose.height=512;
  evt.xexpose.display=dpy;
  evt.xexpose.window=win;
  evt.xexpose.count=0;
  XSendEvent(dpy,win,True,ExposureMask,&evt);
  XFlush(dpy);
}

void *
input_main(void *garbage)
{
  XtAppMainLoop(app_context);
}
