/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File lgdlib.c */
/* This file contains the Legendes basic portable library, in other
 * words, small, portable wrappers around the library functions and
 * system calls that are not as portable as one would want them to be;
 * also, the functions defined here have a die-on-failure option so we
 * don't have to check all return codes all the time. */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
/* XXX - Please insert the appropriate #ifdef's for the following
 * lines. */
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>

#include "lgd0.h"
#include "config.h"
#include "lgdlib.h"
/* Add as few include files as possible here: we want this to function
 * as a standalone library. */

void
lgd_fatal(char *fmt,...)
     /* Terminate the program with a fatal error. Format for arguments
      * is the same as with printf(). */
{
  va_list ap;
  va_start(ap,fmt);
  fprintf(stderr,"Error: ");
#ifdef HAVE_VPRINTF
  vfprintf(stderr,fmt,ap);
#else
  fprintf(stderr,"%s",fmt); /* That's the best I thought of... */
#endif
  fprintf(stderr,"\n");
  if (errno!=0)
    perror("Error");
  va_end(ap);
  exit(EXIT_FAILURE);
}

void
lgd_free(void *ptr)
     /* Free a block of memory as free() does, but with the explicit
      * specification that free(NULL) does nothing. */
{
  if (ptr==NULL) return;
  free(ptr);
}

void *
lgd_malloc(char die_on_failure,size_t size)
     /* Allocate a block of memory as malloc() does, with the
      * following differences: (1)if size==0, do nothing and return
      * NULL, (2)optionally cause a fatal error if malloc() returns
      * NULL. */
{
  void *rtn;
  if (size==0) return NULL;
  rtn=malloc(size);
  if (die_on_failure&&(rtn==NULL))
    lgd_fatal(MSG_OOM);
  return rtn;
}

void *
lgd_realloc(char die_on_failure,void *ptr,size_t size)
     /* Reallocate a block of memory as realloc() does but with the
      * following differences: (1)if size==0, act as free(), (2)if
      * ptr==NULL, act as malloc(), (3)optionally cause a fatal error
      * if realloc() fails. */
{
  void *rtn;
  if (size==0) {
    lgd_free(ptr);
    return NULL; /* This is _not_ a failure code. */
  }
  if (ptr==NULL)
    return lgd_malloc(die_on_failure,size);
  rtn=realloc(ptr,size);
  if (die_on_failure&&(rtn==NULL))
    lgd_fatal(MSG_OOM);
  return rtn;
}

#ifdef HAVE_DEV_RANDOM
static int dev_random; /* File descriptor of /dev/random */
#else
#ifndef HAVE_GOOD_RANDOM
static unsigned long int randseed;
#endif
#endif

void
lgd_randomize(void)
     /* Initialize the random number generator. */
{
#ifdef HAVE_DEV_RANDOM
  if ((dev_random=open(DEV_RANDOM_PATH,O_RDONLY))==-1)
    lgd_fatal(MSG_CANT_OPEN_DEV_RANDOM);
#else
#ifdef HAVE_GOOD_RANDOM
  srandom(time(NULL));
#else
  randseed=time(NULL);
#endif
#endif
}

static unsigned long int
lgd__random(void)
     /* Return a random number between 0 and some unspecified large
      * number. */
{
#ifdef HAVE_DEV_RANDOM
  unsigned long int data;
  if (read(dev_random,&data,sizeof(unsigned long int))
      <sizeof(unsigned long int))
    lgd_fatal(MSG_CANT_READ_DEV_RANDOM);
  return data;
#else
#if HAVE_GOOD_RANDOM
  return random();
#else
  /* XXX - This is VERY, VERY bad. Change this. */
  randseed=(randseed<<(sizeof(unsigned long int)*CHAR_BITS-3))
    |(randseed>>3);
  return randseed=(randseed*169286975UL)+13;
#endif
#endif
}

int
lgd_random(int n)
     /* Return a random number between 0 and n-1. */
{
  return lgd__random()%n;
}

int
lgd_open(char die_on_failure,lgd_filedesc_t *f,char *name,
	 lgd_open_mode_t open_mode)
     /* Open the file named name and return its descriptor in the
      * object pointed by f. If die_on_failure is set, this may not
      * fail. If open_mode is set to OPEN_READONLY, the file is opened
      * rean-only (it should already exist). If open_mode is set to
      * OPEN_READWRITE, the file is created if it does not already
      * exist and is opened read/write if it does, with the file
      * pointer positioned at the beginning. If open_mode is set to
      * OPEN_TRUNCATE, the file is created if it does not already
      * exist and is truncated to zero size and opened read/write if
      * it does. Return value is <0 in case of failure. */
{
#ifdef HAVE_LOWLEVEL_IO
  int fd;
  if (open_mode==OPEN_READONLY) fd=open(name,O_RDONLY);
  else if (open_mode==OPEN_READWRITE) fd=open(name,O_RDWR|O_CREAT,0666);
  else if (open_mode==OPEN_TRUNCATE) fd=open(name,O_RDWR|O_TRUNC|O_CREAT,0666);
  else abort();
  if (die_on_failure&&(fd<0))
    lgd_fatal(MSG_CANT_OPEN_FILE,name);
  *f=fd;
  return (fd<0 ? -1 : 0);
#else
  FILE *fd;
  if (open_mode==OPEN_READONLY) fd=fopen(name,"r");
  else if (open_mode==OPEN_READWRITE) fd=fopen(name,"a+");
  else if (open_mode==OPEN_TRUNCATE) fd=fopen(name,"w+");
  else abort();
  if (die_on_failure&&(fd==NULL))
    lgd_fatal(MSG_CANT_OPEN_FILE,name);
  if (open_mode==OPEN_READWRITE) /* "a+" sets pointer to EOF: we move. */
    rewind(fd);
  *f=fd;
  return (fd==NULL ? -1 : 0);
#endif
}

int
lgd_close(char die_on_failure,lgd_filedesc_t f)
     /* Close the given file descriptor. If die_on_failure is set,
      * this is not allowed to fail. Return value is <0 in case of
      * failure. */
{
  int rtn;
#ifdef HAVE_LOWLEVEL_IO
  rtn=close(f);
  if (die_on_failure&&(rtn<0))
    lgd_fatal(MSG_CANT_CLOSE_FILE);
  return rtn;
#else
  rtn=fclose(f);
  if (die_on_failure&&(rtn==EOF))
    lgd_fatal(MSG_CANT_CLOSE_FILE);
  return (rtn==EOF ? -1 : 0);
#endif
}

int
lgd_seek(char die_on_failure,lgd_filedesc_t f,off_t offset,int whence)
     /* Seek offset in file f, starting from the beginning of the
      * file, the current position or the end of the file according as
      * whence is SEEK_SET, SEEK_CUR or SEEK_END. If die_on_failure is
      * set, this may not fail. Return <0 in case of failure. */
{
#ifdef HAVE_LOWLEVEL_IO
  off_t rtn;
  rtn=lseek(f,offset,whence);
  if (die_on_failure&&(rtn<0))
    lgd_fatal(MSG_CANT_SEEK_FILE);
  return (rtn<0 ? -1 : 0);
#else
  int rtn;
  rtn=fseek(f,offset,whence);
  if (die_on_failure&&(rtn<0))
    lgd_fatal(MSG_CANT_SEEK_FILE);
  return rtn;
#endif
}

off_t
lgd_tell(char die_on_failure,lgd_filedesc_t f)
     /* Return file pointer position for f, measured in bytes from the
      * beginning of the file. If die_on_failure is set, this may not
      * fail. Return <0 in case of failure. */
{
  off_t rtn;
#ifdef HAVE_LOWLEVEL_IO
  rtn=lseek(f,0L,SEEK_CUR);
  if (die_on_failure&&(rtn<0))
    lgd_fatal(MSG_CANT_SEEK_FILE);
  return rtn;
#else
  rtn=ftell(f);
  if (die_on_failure&&(rtn<0))
    lgd_fatal(MSG_CANT_SEEK_FILE);
  return rtn;
#endif
}

ssize_t
lgd_size(char die_on_failure,lgd_filedesc_t f)
     /* Return the file size. If die_on_failure is set, this may not
      * fail. Return <0 in case of failure. */
{
  off_t save;
  ssize_t siz;
  if ((save=lgd_tell(die_on_failure,f))<0)
    return -1;
  if (lgd_seek(die_on_failure,f,0,SEEK_END)<0)
    return -1;
  if ((siz=lgd_tell(die_on_failure,f))<0)
    return -1;
  if (lgd_seek(die_on_failure,f,save,SEEK_SET)<0)
    return -1;
  return siz;
}

ssize_t
lgd_read(char die_on_failure,lgd_filedesc_t f,void *buf,size_t count)
     /* Read count bytes from file f and store them in buf. If
      * die_on_failure is set, this may not fail (reading less than
      * count bytes is considered a failure). Return the number of
      * bytes read, <0 in case of failure. */
{
  ssize_t rtn;
#ifdef HAVE_LOWLEVEL_IO
  rtn=read(f,buf,count);
  if (die_on_failure&&(rtn<count))
    lgd_fatal(MSG_CANT_READ_FILE);
  return rtn;
#else
  rtn=fread(buf,1,count,f);
  if (die_on_failure&&(rtn<count))
    lgd_fatal(MSG_CANT_READ_FILE);
  return (ferror(f) ? -1 : rtn); /* XXX - Not very good. */
#endif
}

ssize_t
lgd_write(char die_on_failure,lgd_filedesc_t f,void *buf,size_t count)
     /* Write count bytes from buf to file f. If die_on_failure is
      * set, this may not fail (writing less than count bytes is
      * considered a failure). Return the number of bytes written, <0
      * in case of failure. */
{
  ssize_t rtn;
#ifdef HAVE_LOWLEVEL_IO
  rtn=write(f,buf,count);
  if (die_on_failure&&(rtn<count))
    lgd_fatal(MSG_CANT_WRITE_FILE);
  return rtn;
#else
  rtn=fwrite(buf,1,count,f);
  if (die_on_failure&&(rtn<count))
    lgd_fatal(MSG_CANT_WRITE_FILE);
  return (ferror(f) ? -1 : rtn); /* XXX - Not very good. */
#endif
}

#ifdef WORDS_BIGENDIAN
static void
swap_int32(u_int32_t *n)
{
  int i;
  char buf1[sizeof(u_int32_t)];
  char buf2[sizeof(u_int32_t)];
  u_int32_t *n1,*n2;
  n1=(u_int32_t *)buf1;
  *n1=*n;
  for (i=0;i<sizeof(u_int32_t);i++)
    buf2[i]=buf1[sizeof(u_int32_t)-i-1];
  n2=(u_int32_t *)buf2;
  *n=*n2;
}
#endif

ssize_t
lgd_read_littleendian_int32(char die_on_failure,lgd_filedesc_t f,
			    u_int32_t *buf,off_t off,size_t count)
     /* Read count 32-bit words in little-endian format from file f,
      * starting from offset off (measured in 32-bit units) and store
      * them in buf.  If die_on_failure is set, this may not fail.
      * Return number of words read, <0 in case of failure. */
{
  ssize_t read;
#ifdef WORDS_BIGENDIAN
  ssize_t i; /* What should this type be? */
#endif
  if (lgd_seek(die_on_failure,f,off*sizeof(u_int32_t),SEEK_SET)<0)
    return -1;
  read=lgd_read(die_on_failure,f,buf,count*sizeof(u_int32_t));
  if (read<0) return -1;
  read/=sizeof(u_int32_t);
#ifdef WORDS_BIGENDIAN
  for (i=0;i<read;i++)
    swap_int32(buf+i);
#endif
  return read;
}

ssize_t
lgd_write_littleendian_int32(char die_on_failure,lgd_filedesc_t f,
			    u_int32_t *buf,off_t off,size_t count)
     /* Write count 32-bit words in little-endian format from buf to
      * file f, starting from offset off (measured in 32-bit units).
      * If die_on_failure is set, this may not fail.  Return number of
      * words written, <0 in case of failure. */
     /* XXX - This function is FAR, FAR too complicated.  It should be
      * possible to write something this trivial in just three lines.
      * Would somebody please do that? */
{
#ifndef WORDS_BIGENDIAN
  ssize_t written;
  if (lgd_seek(die_on_failure,f,off*sizeof(u_int32_t),SEEK_SET)<0)
    return -1;
  written=lgd_write(die_on_failure,f,buf,count*sizeof(u_int32_t));
  if (written<0) return -1;
  written/=sizeof(u_int32_t);
  return written;
#else
  ssize_t written,jw;
  ssize_t i,ptr;
  u_int32_t block[1024];
  if (lgd_seek(die_on_failure,f,off*sizeof(u_int32_t),SEEK_SET)<0)
    return -1;
  written=0;
  while (1) {
    ptr=written;
    for (i=0;(i<1024)&&(ptr<count);i++) {
      block[i]=buf[ptr++];
      swap_int32(block+i);
    }
    if (i==0) return written;
    jw=lgd_write(die_on_failure,f,block,i*sizeof(u_int32_t));
    if (jw>=0) {
      jw/=sizeof(u_int32_t);
      written+=jw;
    }
    if (jw<i*sizeof(u_int32_t)) {
      if (written==0) return -1;
      return written;
    }
  }
#endif
}
